# Copyright 2017 6WIND S.A.

import os
import re
import shutil
import ssl
import subprocess
import urllib

try:
    from pylxd import network, container, Client
except ImportError:
    from pylxd import Client
    from pylxd.models import network, container
from distutils.version import LooseVersion

import charmhelpers.contrib.openstack.utils as ch_utils
import charmhelpers.contrib.openstack.neutron as neutron
import charmhelpers.core.hookenv as hookenv
import charmhelpers.core.files as files
import charmhelpers.core.host as host
import charmhelpers.fetch as fetch

class VirtualAcceleratorCharmTools():

    # Internal name of charm - used for HA support + others
    name = 'va'

    # supported versions
    va_versions = [
        '1.2',
        '1.3',
        '1.4',
        '1.5',
        '1.6',
        '1.7'
    ]

    # First release of openstack this charm supports
    release = 'liberty'

    # Packages the service needs installed
    packages = [
        'libcurl3',
        'libjansson4',
        neutron.headers_package()
    ]

    resource_creds = 'credentials'
    resource_va_license = 'license'
    product = 'virtual-accelerator'
    ovs_pkg = 'openvswitch-switch'

    license_file = '/etc/va.lic'
    startup_script = '/usr/bin/%s.sh' % product
    cpuset_env = '/etc/cpuset.env'

    def __init__(self, **kwargs):
        version = hookenv.config('va-version')
        if version in ['1.2', '1.3', '1.4']:
            self.license_file = '/usr/local/etc/va.lic'
            self.startup_script = '/usr/local/bin/%s.sh' % self.product
            self.cpuset_env = '/usr/local/etc/cpuset.env'

    def _get_dpkg_arch(self):
        """Returns the running machine's architecture.
        """
        return subprocess.check_output(['dpkg', '--print-architecture'],
                                       universal_newlines=True).rstrip()

    def _get_current_va_version(self):
        """Get product debian package version.
        """
        apt = subprocess.Popen(['apt-cache', 'policy', 'virtual-accelerator'],
                               stdout=subprocess.PIPE)
        grep = subprocess.Popen(['grep', 'Installed'], stdin=apt.stdout,
                                stdout=subprocess.PIPE)
        apt.stdout.close()
        version = re.search('([0-9.]+[0-9])',
                            grep.communicate()[0].decode('ascii')).group(1)
        return version

    def _install_deb(self, pkg):
        """Install the provided debian package.
        """
        return subprocess.check_output(['dpkg', '-i', pkg])

    def _fetch_6wind_repo_pkg(self, output, version):
        """Download the 6WIND package providing 6WIND repository.
        It requires an SSL connection to 6WIND servers through an SSL context.
        Put the package into the output file.
        """
        distrib = 'ubuntu-{}'.format(ch_utils.lsb_release()['DISTRIB_RELEASE'])
        arch = self._get_dpkg_arch()

        pkg = '6wind-%s-%s-repository_%s-1_%s.deb' % \
            (self.product, distrib, version, arch)
        subdir = '%s/%s/%s/%s' % (self.product, distrib, arch, version)
        repo_url = 'https://repo.6wind.com/%s/%s' % (subdir, pkg)

        certs_dir = '/usr/local/etc/certs'
        cacert = certs_dir + '/6wind_ca.crt'
        cert = certs_dir + '/6wind_client.crt'
        key = certs_dir + '/6wind_client.key'

        ctx = ssl.create_default_context(purpose=ssl.Purpose.CLIENT_AUTH,
                                         cafile=cacert)
        ctx.load_cert_chain(cert, key)

        hookenv.log('Using the following URL for 6WIND repo: %s' % \
                    repo_url, level=hookenv.INFO)
        response = urllib.request.urlopen(repo_url, context=ctx)

        host.write_file(output, response.read())

        return

    def _remove_6wind_repo(self):
        """Remove the 6WIND repository package.
        """
        hookenv.status_set('maintenance', 'Uninstalling 6WIND repository')
        hookenv.log('Uninstalling 6WIND repository', level=hookenv.INFO)
        distrib = 'ubuntu-{}'.format(ch_utils.lsb_release()['DISTRIB_RELEASE'])
        arch = self._get_dpkg_arch()

        pkg = '6wind-%s-%s-repository' % (self.product, distrib)
        fetch.apt_purge(pkg)
        fetch.apt_update()

    def install_prerequisites(self):
        """Make sure that the correct linux headers are installed.
        For xenial and later, use distro packages since patches are upstream.
        """
        hookenv.status_set('maintenance', 'Installing prerequisites')
        hookenv.log('Installing linux headers',
                    level=hookenv.INFO)
        fetch.apt_install(self.packages,
                          options=['--option=Dpkg::Options::=--force-confnew'])

    def install_credentials(self):
        """Install 6WIND credentials using apt, to enable 6WIND repository
        access.
        """
        hookenv.status_set('maintenance', 'Installing credentials')
        resource = hookenv.resource_get(self.resource_creds)
        if not resource:
            hookenv.log('Missing required credentials resource',
                        level=hookenv.CRITICAL)
            raise Exception('Missing credentials resource: '
                            '{}'.format(self.resource_creds))

        hookenv.log('Installing 6WIND credentials', level=hookenv.INFO)
        self._install_deb(resource)

    def __post_install(self):
        """Do some post-install voodoo.
        """
        hookenv.status_set('maintenance', 'Setting up LXC netns sync with vrf')
        files.sed(self.startup_script, 'vrfd -s', 'vrfd -ls')

        hookenv.status_set('maintenance', 'Disabling CPUSET')
        files.sed(self.cpuset_env, '.*CPUSET_ENABLE.*', 'CPUSET_ENABLE=0')

        hookenv.status_set('maintenance',
                           'Moving %s.service to multi-user systemd unit'
                           % self.product)
        # let's start VA within multi-user.target unit, otherwise it is often
        # in a race conditions with cloud-init.service, which is in the same
        # network-pre.target as VA and calls ifup@.service slightly at the same
        # moment, when fast-path.sh performs its "if down/up".
        subprocess.Popen(['sudo', 'cp', 'rc-local.service',
                          '/etc/systemd/system'], stdout=subprocess.PIPE)
        subprocess.Popen(['sudo', 'rm', '/usr/lib/systemd/system/%s.service'
                          % self.product], stdout=subprocess.PIPE)
        subprocess.Popen(['sudo', 'ln', '-s',
                          '/etc/systemd/system/rc-local.service',
                          '/usr/lib/systemd/system/%s.service' % self.product],
                          stdout=subprocess.PIPE)
        host.service('enable', self.product)
        # to avoid dpkg --configure error, due to potential openvswitch-switch
        # package downgrade
        host.service('enable', 'openvswitch-switch')
        # to avoid systemd warn message, that rc-local.service changed on disk
        subprocess.Popen(['sudo', 'systemctl', 'daemon-reload'],
                         stdout=subprocess.PIPE)

    def _install_6wind_repos_pkg(self, va_version):
        """Install 6WIND repository package and update database.
        """
        repo_pkg = '/tmp/6wind-va-%s-repo.deb' % va_version

        hookenv.log('Downloading 6WIND %s repo package' % self.product,
                    level=hookenv.INFO)
        self._fetch_6wind_repo_pkg(repo_pkg, va_version)

        hookenv.log('Installing 6WIND %s repo package' % self.product,
                    level=hookenv.INFO)
        self._install_deb(repo_pkg)

        fetch.apt_update()

    def install_va(self):
        """Install the product using the 6WIND repository.
        While at it, mark OVS on hold, to prevent a version without 6WIND
        patches from being installed on upgrade.
        """
        hookenv.status_set('maintenance', 'Installing %s' % self.product)

        version = hookenv.config('va-version')
        self._install_6wind_repos_pkg(version)

        hookenv.log('Installing 6WIND %s' % self.product, level=hookenv.INFO)
        fetch.apt_install(self.product, options=['--allow-downgrades'])

        self.__post_install()

        hookenv.log('Set hold mark on %s to prevent our 6WIND version from being'
                    ' overridden' % self.ovs_pkg, level=hookenv.INFO)
        fetch.apt_hold(self.ovs_pkg)

    def upgrade_va(self, va_version):
        """Upgrade the product using the 6WIND repository if the installed
        version is lower than va-version. While at it, mark OVS on hold, to
        prevent a version without 6WIND patches from being installed on
        upgrade.
        """
        hookenv.status_set('maintenance', 'Upgrading %s' % self.product)

        current_version = self._get_current_va_version()
        if LooseVersion(current_version) >= LooseVersion(va_version):
            hookenv.log('Current %s version is greater than or equal to the'
                        'version you want to install, nothing to do' % (self.product),
                        level=hookenv.WARNING)

        hookenv.log('Upgrading %s from v%s -> v%s'
                    %(self.product, current_version, va_version),
                    level=hookenv.INFO)
        self._install_6wind_repos_pkg(va_version)

        hookenv.log('Unhold mark temporary on %s to be able to upgrade 6WIND'
                    '%s' % (self.ovs_pkg, self.product), level=hookenv.INFO)
        fetch.apt_unhold(self.ovs_pkg)

        hookenv.log('Upgrading 6WIND %s' % self.product, level=hookenv.INFO)
        fetch.apt_install(self.product,
                          options=['--only-upgrade',
                                   '--option=Dpkg::Options::=--force-confold'])

        self.__post_install()

        hookenv.log('Set hold mark on %s to prevent our 6WIND version from'
                    'being overridden' % self.ovs_pkg, level=hookenv.INFO)
        fetch.apt_hold(self.ovs_pkg)

    def uninstall_va(self):
        """Uninstall the product, and all packages related to 6WIND.
        Revert the OVS package to the one without 6WIND patches.
        """
        self._remove_6wind_repo()

        hookenv.status_set('maintenance', 'Uninstalling %s' % self.product)
        hookenv.log('Uninstalling %s' % self.product, level=hookenv.INFO)
        fetch.apt_purge(self.product)

        hookenv.status_set('maintenance', 'Reinstall original %s' % self.ovs_pkg)
        hookenv.log('Reinstall original %s' % self.ovs_pkg, level=hookenv.INFO)
        fetch.apt_unhold(self.ovs_pkg)
        cache = fetch.apt_cache()
        pkg = cache[self.ovs_pkg]
        for ver in pkg.version_list:
            # select the closest version in cache, not coming from 6WIND
            if '6wind' in ver.ver_str:
                continue
            fetch.apt_install('%s=%s' % (self.ovs_pkg, ver.ver_str),
                              ['--option=Dpkg::Options::=--force-confold',
                               '--allow-downgrades'])
            break

        host.service_restart(self.ovs_pkg)
        hookenv.status_set('maintenance', 'Final cleanup')
        hookenv.log('Final cleanup', level=hookenv.INFO)
        subprocess.check_output(['apt-get', '--assume-yes', 'autoremove'])
        fetch.apt_purge('6wind*')

    def install_license(self):
        """Put the license resource into the default folder.
        There, il will be scanned when starting the product.
        """
        hookenv.status_set('maintenance', 'Installing license')
        resource = hookenv.resource_get(self.resource_va_license)
        if not resource:
            hookenv.log('VA will run temporarily, without license',
                        level=hookenv.INFO)
            return

        f = open(resource, 'r')
        for line in f:
            if 'BOILERPLATE' in line:
                # ignore this file, it's not a real license file.
                f.close()
                return
        f.close()

        hookenv.log('Installing 6WIND license into ', level=hookenv.INFO)
        shutil.copyfile(resource, self.license_file)
        return

    def uninstall_license(self):
        """Remove the license file from the unit.
        """
        hookenv.status_set('maintenance', 'Uninstalling license')
        try:
            os.remove(self.license_file)
        except FileNotFoundError:
            return

    def _lxc_bridge_ports_readd(self):
        """When VA restarts, it shuts off all interfaces, including the bridge
        containing LXC containers veth. As they are configured dynamically when
        starting the container itself, there's no /etc/network config for them
        and ifup can't bring them up again.
        Here we use pylxd to get the interfaces names on the host and set them
        up again in the correct bridge.
        """
        hookenv.status_set('maintenance', 'Updating LXC bridge ports')

        client = Client()
        networks = network.Network.all(client)

        for net in networks:
            if net.used_by:
                for c in net.used_by:
                    ct = container.Container.get(client, c.split('/')[-1])
                    ct_net = ct.state().network
                    for iface, conf in ct_net.items():
                        if conf['host_name']:
                            host_name = conf['host_name']
                            try:
                                subprocess.check_output(['brctl', 'addif',
                                                         net.name, host_name])
                            except subprocess.CalledProcessError as e:
                                hookenv.log(e.output, level=hookenv.WARNING)
                            hookenv.log('Added %s to bridge %s.' %
                                        (host_name, net.name),
                                        level=hookenv.INFO)

