# Copyright 2017 6WIND S.A.

import libvirt
import os
import re
import shutil
import ssl
import subprocess
import tempfile
import urllib
import fileinput

try:
    from pylxd import network, container, Client
except ImportError:
    from pylxd import Client
    from pylxd.models import network, container

import charmhelpers.contrib.openstack.utils as ch_utils
import charmhelpers.contrib.openstack.neutron as neutron
import charmhelpers.core.hookenv as hookenv
import charmhelpers.core.files as files
import charmhelpers.core.host as host
import charmhelpers.fetch as fetch

from charms_openstack.charm import OpenStackCharm
from charms_openstack.adapters import OpenStackRelationAdapters

# import 6WIND layer
import charm.openstack.utils as va_utils

class VirtualAcceleratorCharm(OpenStackCharm):
    """VirtualAcceleratorCharm defines all possible actions for managing VA.
    """

    # Instanciate Virtual Accelerator Charm Layer class
    va_common = va_utils.VirtualAcceleratorCharmTools()

    # First release of openstack this charm supports
    release = 'liberty'

    # Packages the service needs installed
    packages = [
        'libvirt-bin',
        'python-libvirt',
        'qemu',
        'qemu-system-x86',
        'libcurl3',
        'libjansson4',
        neutron.headers_package()
    ]

    # Standard interface adapters class to use.
    adapters_class = OpenStackRelationAdapters

    # Ports that need exposing.
    default_service = ''

    resource_fp_conf = 'custom_fp_conf'
    product = 'virtual-accelerator'
    os_extensions = '6wind-openstack-extensions'

    fp_config = '/etc/fast-path.env'
    fpm_config = '/etc/fpm.env'
    cmgr_config = '/etc/cmgr.env'
    license_file = '/etc/va.lic'
    startup_script = '/usr/bin/%s.sh' % product
    cpuset_env = '/etc/cpuset.env'

    def ports_to_check(self, ports):
        """VA does not expose ports. Don't check any ports automatically.

        Refer to:
        https://github.com/openstack/charms.openstack#not-checking-services-are-running
        """
        return []

    def __init__(self, release=None, **kwargs):
        """Custom initialiser for class
        If no release is passed, then the charm determines the release from the
        ch_utils.os_release() function.
        """
        version = hookenv.config('va-version')
        if version in ['1.2', '1.3', '1.4']:
            self.fp_config = '/usr/local/etc/fast-path.env'
            self.license_file = '/usr/local/etc/va.lic'
            self.startup_script = '/usr/local/bin/%s.sh' % self.product
            self.cpuset_env = '/usr/local/etc/cpuset.env'

        if release is None:
            release = ch_utils.os_release('nova-common')
        OpenStackCharm.__init__(self, release=release, **kwargs)

    def _remove_6wind_repo(self):
        """Remove the 6WIND repository package.
        """
        hookenv.status_set('maintenance', 'Uninstalling 6WIND repository')
        hookenv.log('Uninstalling 6WIND repository', level=hookenv.INFO)
        distrib = 'ubuntu-{}'.format(ch_utils.lsb_release()['DISTRIB_RELEASE'])
        arch = self._get_dpkg_arch()

        pkg = '6wind-%s-%s-repository' % (self.product, distrib)
        fetch.apt_purge(pkg)
        fetch.apt_update()

    def install_prerequisites(self):
        """Install correct libvirt and qemu packages.
        For xenial and later, use distro packages since patches are upstream.
        """
        hookenv.status_set('maintenance', 'Installing prerequisites')
        hookenv.log('Installing libvirt, qemu and linux headers',
                    level=hookenv.INFO)
        fetch.apt_install(self.packages,
                          options=['--option=Dpkg::Options::=--force-confnew'])

    def delete_default_libvirt_net(self):
        """Remove the 'default' libvirt network as it otherwise sets up
        bridges and complicated iptables rules, that impedes readability.
        """
        hookenv.status_set('maintenance', 'Cleaning libvirt default network')
        conn = libvirt.open()
        try:
            default_net = conn.networkLookupByName('default')
        except libvirt.libvirtError:
            return conn.close()

        hookenv.log('Deleting default libvirt network', level=hookenv.INFO)
        if default_net.isActive():
            default_net.destroy()
        default_net.undefine()

    def install_os_extensions(self):
        """Install OpenStack extensions that provide 6WIND monkey patch.
        """
        hookenv.status_set('maintenance', 'Installing 6WIND Openstack Extensions')
        hookenv.log('Installing 6WIND Openstack Extensions', level=hookenv.INFO)
        fetch.apt_install(self.os_extensions)

    def upgrade_os_extensions(self):
        """Upgrade OpenStack extensions.
        """
        hookenv.status_set('maintenance', 'Upgrading 6WIND Openstack Extensions')
        hookenv.log('Upgrading 6WIND Openstack Extensions', level=hookenv.INFO)
        fetch.apt_install(self.os_extensions, options=['--only-upgrade'])

    def uninstall_os_extensions(self):
        """Uninstall OpenStack extensions.
        """
        hookenv.status_set('maintenance', 'Uninstalling 6WIND Openstack Extensions')
        hookenv.log('Uninstalling 6WIND Openstack Extensions', level=hookenv.INFO)
        fetch.apt_purge(self.os_extensions)

    def render_config(self):
        """Use the FP Wizard to generate the fast path config file, considering
        the charm's configuration options.
        """
        hookenv.status_set('maintenance', 'Generating %s' % self.fp_config)
        hookenv.log('Generating %s' % self.fp_config, level=hookenv.INFO)

        tmp_file = None
        resource = None
        try:
            resource = hookenv.resource_get(self.resource_fp_conf)
        except Exception:
            hookenv.log('Custom fp configuration file not found, using juju charm config...',
                        level=hookenv.INFO)
        if resource:
            with open(resource, 'r') as f:
                line = f.readline()
                if 'BOILERPLATE' not in line:
                    hookenv.log('Use provided custom fp config file',
                                level=hookenv.INFO)
                    tmp_file = resource
                else:
                    hookenv.log('Boilerplate fp config. Using default conf.',
                                level=hookenv.INFO)
                f.close()

        if not tmp_file:
            tmp_conf = tempfile.NamedTemporaryFile()
            vm_mem = hookenv.config('VM_MEMORY')
            tmp_conf.write(str.encode('VM_MEMORY=%s\n' % vm_mem))
            fp_mask = hookenv.config('FP_MASK')
            tmp_conf.write(str.encode('FP_MASK=%s\n' % fp_mask))
            tmp_conf.flush()
            tmp_file = tmp_conf.name

        subprocess.run(['fp-conf-tool', '--update', '--file=%s' % tmp_file])

        shutil.copyfile(tmp_file, self.fp_config)

        if tmp_conf:
            tmp_conf.close()

        if hookenv.config('DEBUG_MODE') == "on":
            with fileinput.FileInput(self.cmgr_config, inplace=True) as cmgr:
                for line in cmgr:
                    line = re.sub(r"^#:\s\${CMGR_OPTIONS:=.*}",
                                  ": ${CMGR_OPTIONS:=\"-d 0xffffffff\"}",
                                  line.rstrip())
                    print (line)
                cmgr.close()
            with fileinput.FileInput(self.fpm_config, inplace=True) as fpm:
                for line in fpm:
                    line = re.sub(r"^#:\s\${FPM_OPTIONS:=.*}",
                                  ": ${FPM_OPTIONS:=\"-v\"}",
                                  line.rstrip())
                    print (line)
                fpm.close()
        else:
            with fileinput.FileInput(self.cmgr_config, inplace=True) as cmgr:
                for line in cmgr:
                    line = re.sub(r"^:\s\${CMGR_OPTIONS:=.*}",
                                  "#: ${CMGR_OPTIONS:=\"-d 0xffffffff\"}",
                                  line.rstrip())
                    print (line)
                cmgr.close()
            with fileinput.FileInput(self.fpm_config, inplace=True) as fpm:
                for line in fpm:
                    line = re.sub(r"^:\s\${FPM_OPTIONS:=.*}",
                                  "#: ${FPM_OPTIONS:=\"-v\"}",
                                  line.rstrip())
                    print (line)
                fpm.close()

    def _dependencies_restart(self):
        """Restart all dependent on services required to work properly.
        """
        services = [
            'libvirt-bin',
            'openvswitch-switch',
            'neutron-openvswitch-agent'
        ]
        for service in services:
            hookenv.status_set('maintenance', 'Restarting %s' % service)
            hookenv.log('Restarting %s' % service, level=hookenv.INFO)
            host.service_restart(service)
        # This may not look like a service but let's do it here to keep just
        # one clean method for all dependancies
        self.va_common._lxc_bridge_ports_readd()

    def restart(self):
        """Restart the product, typically after changing configuration options.
        Restart libvirt afterwards, to take fast path hugepages into account.
        """
        hookenv.status_set('maintenance', 'Restarting %s' % self.product)
        hookenv.log('Restarting %s' % self.product, level=hookenv.INFO)
        host.service_restart(self.product)

        self._dependencies_restart()

    def stop(self):
        """Stop the running product service. This is done when terminating.
        """
        hookenv.status_set('maintenance', 'Stopping %s' % self.product)
        hookenv.log('Stopping %s' % self.product, level=hookenv.INFO)
        host.service_stop(self.product)

        self._dependencies_restart()

    def start(self):
        """Start product service. This is called when finishing installation.
        """
        hookenv.status_set('maintenance', 'Starting %s' % self.product)
        hookenv.log('Starting %s...' % self.product, level=hookenv.INFO)
        host.service_start(self.product)

        self._dependencies_restart()
